@@ -24,7 +24,7 @@ __copyright__ = '(C) 2016, Arnaud Morvan'
 
 # This will get replaced with a git SHA1 when you do a git archive
 
-__revision__ = '8a899c8758b30adc3f8d6df682a5702dd3bfe6f0'
+__revision__ = '9b176802e5512bc310765d8ebe4e44b5f4663453'
 
 import locale
 import os
@@ -34,6 +34,7 @@ from copy import deepcopy
 
 from qgis.core import (
     QgsApplication,
+    QgsUnitTypes,
     QgsCoordinateReferenceSystem,
     QgsExpression,
     QgsExpressionContextGenerator,
@@ -53,6 +54,7 @@ from qgis.core import (
     QgsProcessingParameterFile,
     QgsProcessingParameterMultipleLayers,
     QgsProcessingParameterNumber,
+    QgsProcessingParameterDistance,
     QgsProcessingParameterRasterLayer,
     QgsProcessingParameterEnum,
     QgsProcessingParameterString,
@@ -62,6 +64,8 @@ from qgis.core import (
     QgsProcessingParameterFeatureSource,
     QgsProcessingParameterMapLayer,
     QgsProcessingParameterBand,
+    QgsProcessingParameterMatrix,
+    QgsProcessingParameterDistance,
     QgsProcessingFeatureSourceDefinition,
     QgsProcessingOutputRasterLayer,
     QgsProcessingOutputVectorLayer,
@@ -77,6 +81,7 @@ from qgis.core import (
 from qgis.PyQt.QtWidgets import (
     QCheckBox,
     QComboBox,
+    QLabel,
     QDialog,
     QFileDialog,
     QHBoxLayout,
@@ -102,7 +107,7 @@ from qgis.utils import iface
 from processing.core.ProcessingConfig import ProcessingConfig
 from processing.modeler.MultilineTextPanel import MultilineTextPanel
 
-from processing.gui.NumberInputPanel import NumberInputPanel, ModellerNumberInputPanel
+from processing.gui.NumberInputPanel import NumberInputPanel, ModelerNumberInputPanel, DistanceInputPanel
 from processing.gui.RangePanel import RangePanel
 from processing.gui.PointSelectionPanel import PointSelectionPanel
 from processing.gui.FileSelectionPanel import FileSelectionPanel
@@ -150,6 +155,7 @@ class WidgetWrapper(QObject):
         self.col = col
         self.dialogType = dialogTypes.get(dialog.__class__.__name__, DIALOG_STANDARD)
         self.widget = self.createWidget(**kwargs)
+        self.label = self.createLabel()
         if param.defaultValue() is not None:
             self.setValue(param.defaultValue())
 
@@ -172,6 +178,21 @@ class WidgetWrapper(QObject):
     def createWidget(self, **kwargs):
         pass
 
+    def createLabel(self):
+        if self.dialogType == DIALOG_BATCH:
+            return None
+        desc = self.param.description()
+        if isinstance(self.param, QgsProcessingParameterExtent):
+            desc += self.tr(' (xmin, xmax, ymin, ymax)')
+        if isinstance(self.param, QgsProcessingParameterPoint):
+            desc += self.tr(' (x, y)')
+        if self.param.flags() & QgsProcessingParameterDefinition.FlagOptional:
+            desc += self.tr(' [optional]')
+
+        label = QLabel(desc)
+        label.setToolTip(self.param.name())
+        return label
+
     def setValue(self, value):
         pass
 
@@ -240,6 +261,12 @@ class BasicWidgetWrapper(WidgetWrapper):
 
 class BooleanWidgetWrapper(WidgetWrapper):
 
+    def createLabel(self):
+        if self.dialogType == DIALOG_STANDARD:
+            return None
+        else:
+            return super().createLabel()
+
     def createWidget(self):
         if self.dialogType == DIALOG_STANDARD:
             return QCheckBox()
@@ -319,6 +346,7 @@ class CrsWidgetWrapper(WidgetWrapper):
             else:
                 widget.setOptionVisible(QgsProjectionSelectionWidget.CrsNotSet, True)
 
+            widget.crsChanged.connect(lambda: self.widgetValueHasChanged.emit(self))
             return widget
 
     def selectProjection(self):
@@ -529,19 +557,26 @@ class FileWidgetWrapper(WidgetWrapper):
 class FixedTableWidgetWrapper(WidgetWrapper):
 
     def createWidget(self):
-        return FixedTablePanel(self.param)
+        if self.dialogType in (DIALOG_STANDARD, DIALOG_BATCH):
+            return FixedTablePanel(self.param)
+        else:
+            self.combobox = QComboBox()
+            values = self.dialog.getAvailableValuesOfType(QgsProcessingParameterMatrix)
+            for v in values:
+                self.combobox.addItem(self.dialog.resolveValueDescription(v), v)
+            return self.combobox
 
     def setValue(self, value):
-        self.widget.setValue(value)
+        if self.dialogType in (DIALOG_STANDARD, DIALOG_BATCH):
+            self.widget.setValue(value)
+        else:
+            self.setComboValue(value, combobox=self.combobox)
 
     def value(self):
-        if self.dialogType == DIALOG_MODELER:
-            table = self.widget.table
-            if not bool(table) and not self.param.flags() & QgsProcessingParameterDefinition.FlagOptional:
-                raise InvalidParameterValue()
-            return ParameterFixedTable.tableToString(table)
-        else:
+        if self.dialogType in (DIALOG_STANDARD, DIALOG_BATCH):
             return self.widget.table
+        else:
+            return self.comboValue(combobox=self.combobox)
 
 
 class MultipleLayerWidgetWrapper(WidgetWrapper):
@@ -653,7 +688,7 @@ class MultipleLayerWidgetWrapper(WidgetWrapper):
         if self.dialogType == DIALOG_STANDARD:
             pass  # TODO
         elif self.dialogType == DIALOG_BATCH:
-            return self.widget.setText(value)
+            return self.widget.setValue(value)
         else:
             options = self._getOptions()
 
@@ -688,7 +723,7 @@ class MultipleLayerWidgetWrapper(WidgetWrapper):
                                                                         False)
                 return [options[i] if isinstance(i, int) else i for i in self.widget.selectedoptions]
         elif self.dialogType == DIALOG_BATCH:
-            return self.widget.getText()
+            return self.widget.value()
         else:
             options = self._getOptions()
             values = [options[i] if isinstance(i, int) else QgsProcessingModelChildParameterSource.fromStaticValue(i)
@@ -706,7 +741,7 @@ class NumberWidgetWrapper(WidgetWrapper):
             widget.hasChanged.connect(lambda: self.widgetValueHasChanged.emit(self))
             return widget
         else:
-            return ModellerNumberInputPanel(self.param, self.dialog)
+            return ModelerNumberInputPanel(self.param, self.dialog)
 
     def setValue(self, value):
         if value is None or value == NULL:
@@ -729,6 +764,42 @@ class NumberWidgetWrapper(WidgetWrapper):
         self.widget.setDynamicLayer(wrapper.value())
 
 
+class DistanceWidgetWrapper(WidgetWrapper):
+
+    def createWidget(self):
+        if self.dialogType in (DIALOG_STANDARD, DIALOG_BATCH):
+            widget = DistanceInputPanel(self.param)
+            widget.hasChanged.connect(lambda: self.widgetValueHasChanged.emit(self))
+            return widget
+        else:
+            return ModelerNumberInputPanel(self.param, self.dialog)
+
+    def setValue(self, value):
+        if value is None or value == NULL:
+            return
+
+        self.widget.setValue(value)
+
+    def value(self):
+        return self.widget.getValue()
+
+    def postInitialize(self, wrappers):
+        if self.dialogType in (DIALOG_STANDARD, DIALOG_BATCH):
+            for wrapper in wrappers:
+                if wrapper.param.name() == self.param.dynamicLayerParameterName():
+                    self.widget.setDynamicLayer(wrapper.value())
+                    wrapper.widgetValueHasChanged.connect(self.dynamicLayerChanged)
+                if wrapper.param.name() == self.param.parentParameterName():
+                    self.widget.setUnitParameterValue(wrapper.value())
+                    wrapper.widgetValueHasChanged.connect(self.parentParameterChanged)
+
+    def dynamicLayerChanged(self, wrapper):
+        self.widget.setDynamicLayer(wrapper.value())
+
+    def parentParameterChanged(self, wrapper):
+        self.widget.setUnitParameterValue(wrapper.value())
+
+
 class RangeWidgetWrapper(WidgetWrapper):
 
     def createWidget(self):
@@ -737,7 +808,7 @@ class RangeWidgetWrapper(WidgetWrapper):
             widget.hasChanged.connect(lambda: self.widgetValueHasChanged.emit(self))
             return widget
         #else:
-        #    return ModellerNumberInputPanel(self.param, self.dialog)
+        #    return ModelerNumberInputPanel(self.param, self.dialog)
 
     def setValue(self, value):
         if value is None or value == NULL:
@@ -793,10 +864,10 @@ class MapLayerWidgetWrapper(WidgetWrapper):
             self.combo = QComboBox()
             layers = self.getAvailableLayers()
             self.combo.setEditable(True)
+            if self.param.flags() & QgsProcessingParameterDefinition.FlagOptional:
+                self.combo.addItem(self.NOT_SELECTED, self.NOT_SET_OPTION)
             for layer in layers:
                 self.combo.addItem(self.dialog.resolveValueDescription(layer), layer)
-            if self.param.flags() & QgsProcessingParameterDefinition.FlagOptional:
-                self.combo.setEditText("")
 
             widget = QWidget()
             layout = QHBoxLayout()
@@ -830,23 +901,36 @@ class MapLayerWidgetWrapper(WidgetWrapper):
                 self.combo.setCurrentIndex(self.combo.findText(filename))
             else:
                 self.combo.setEditText(filename)
+            self.widgetValueHasChanged.emit(self)
 
     def setValue(self, value):
         if value is None or value == NULL:
             return
 
         if self.dialogType == DIALOG_STANDARD:
-            if self.combo.findText(value) >= 0:
-                self.combo.setCurrentIndex(self.combo.findText(value))
-            else:
-                items = self.combo.additionalItems()
-                items.append(value)
-                self.combo.setAdditionalItems(items)
-                self.combo.setCurrentIndex(self.combo.findText(value))
+            if isinstance(value, str):
+                layer = QgsProject.instance().mapLayer(value)
+                if layer is not None:
+                    value = layer
+
+            found = False
+            if isinstance(value, QgsMapLayer):
+                self.combo.setLayer(value)
+                found = self.combo.currentIndex() != -1
+
+            if not found:
+                if self.combo.findText(value) >= 0:
+                    self.combo.setCurrentIndex(self.combo.findText(value))
+                else:
+                    items = self.combo.additionalItems()
+                    items.append(value)
+                    self.combo.setAdditionalItems(items)
+                    self.combo.setCurrentIndex(self.combo.findText(value))
         elif self.dialogType == DIALOG_BATCH:
-            self.widget.setText(value)
+            self.widget.setValue(value)
         else:
             self.setComboValue(value, combobox=self.combo)
+        self.widgetValueHasChanged.emit(self)
 
     def value(self):
         if self.dialogType == DIALOG_STANDARD:
@@ -891,45 +975,64 @@ class RasterWidgetWrapper(MapLayerWidgetWrapper):
                 self.combo.setCurrentIndex(self.combo.findText(filename))
             else:
                 self.combo.setEditText(filename)
+            self.widgetValueHasChanged.emit(self)
 
 
 class EnumWidgetWrapper(WidgetWrapper):
+    NOT_SELECTED = '[Not selected]'
 
     def createWidget(self, useCheckBoxes=False, columns=1):
-        self._useCheckBoxes = useCheckBoxes
-        if self._useCheckBoxes and not self.dialogType == DIALOG_BATCH:
-            return CheckboxesPanel(options=self.param.options(),
-                                   multiple=self.param.allowMultiple(),
-                                   columns=columns)
-        if self.param.allowMultiple():
-            return MultipleInputPanel(options=self.param.options())
+        if self.dialogType in (DIALOG_STANDARD, DIALOG_BATCH):
+            self._useCheckBoxes = useCheckBoxes
+            if self._useCheckBoxes and not self.dialogType == DIALOG_BATCH:
+                return CheckboxesPanel(options=self.param.options(),
+                                       multiple=self.param.allowMultiple(),
+                                       columns=columns)
+            if self.param.allowMultiple():
+                return MultipleInputPanel(options=self.param.options())
+            else:
+                widget = QComboBox()
+                for i, option in enumerate(self.param.options()):
+                    widget.addItem(option, i)
+                if self.param.defaultValue():
+                    widget.setCurrentIndex(widget.findData(self.param.defaultValue()))
+                return widget
         else:
-            widget = QComboBox()
+            self.combobox = QComboBox()
+            if self.param.flags() & QgsProcessingParameterDefinition.FlagOptional:
+                self.combobox.addItem(self.NOT_SELECTED, self.NOT_SET_OPTION)
             for i, option in enumerate(self.param.options()):
-                widget.addItem(option, i)
-            if self.param.defaultValue():
-                widget.setCurrentIndex(widget.findData(self.param.defaultValue()))
-            return widget
+                self.combobox.addItem(option, i)
+            values = self.dialog.getAvailableValuesOfType(QgsProcessingParameterEnum)
+            for v in values:
+                self.combobox.addItem(self.dialog.resolveValueDescription(v), v)
+            return self.combobox
 
     def setValue(self, value):
         if value is None or value == NULL:
             return
 
-        if self._useCheckBoxes and not self.dialogType == DIALOG_BATCH:
-            self.widget.setValue(value)
-            return
-        if self.param.allowMultiple():
-            self.widget.setSelectedItems(value)
+        if self.dialogType in (DIALOG_STANDARD, DIALOG_BATCH):
+            if self._useCheckBoxes and not self.dialogType == DIALOG_BATCH:
+                self.widget.setValue(value)
+                return
+            if self.param.allowMultiple():
+                self.widget.setSelectedItems(value)
+            else:
+                self.widget.setCurrentIndex(self.widget.findData(value))
         else:
-            self.widget.setCurrentIndex(self.widget.findData(value))
+            self.setComboValue(value, combobox=self.combobox)
 
     def value(self):
-        if self._useCheckBoxes and not self.dialogType == DIALOG_BATCH:
-            return self.widget.value()
-        if self.param.allowMultiple():
-            return self.widget.selectedoptions
+        if self.dialogType in (DIALOG_STANDARD, DIALOG_BATCH):
+            if self._useCheckBoxes and not self.dialogType == DIALOG_BATCH:
+                return self.widget.value()
+            if self.param.allowMultiple():
+                return self.widget.selectedoptions
+            else:
+                return self.widget.currentData()
         else:
-            return self.widget.currentData()
+            return self.comboValue(combobox=self.combobox)
 
 
 class FeatureSourceWidgetWrapper(WidgetWrapper):
@@ -1047,23 +1150,36 @@ class FeatureSourceWidgetWrapper(WidgetWrapper):
                 self.combo.setCurrentIndex(self.combo.findText(filename))
             else:
                 self.combo.setEditText(filename)
+            self.widgetValueHasChanged.emit(self)
 
     def setValue(self, value):
         if value is None or value == NULL:
             return
 
         if self.dialogType == DIALOG_STANDARD:
-            if self.combo.findText(value) >= 0:
-                self.combo.setCurrentIndex(self.combo.findText(value))
-            else:
-                items = self.combo.additionalItems()
-                items.append(value)
-                self.combo.setAdditionalItems(items)
-                self.combo.setCurrentIndex(self.combo.findText(value))
+            if isinstance(value, str):
+                layer = QgsProject.instance().mapLayer(value)
+                if layer is not None:
+                    value = layer
+
+            found = False
+            if isinstance(value, QgsMapLayer):
+                self.combo.setLayer(value)
+                found = self.combo.currentIndex() != -1
+
+            if not found:
+                if self.combo.findText(value) >= 0:
+                    self.combo.setCurrentIndex(self.combo.findText(value))
+                else:
+                    items = self.combo.additionalItems()
+                    items.append(value)
+                    self.combo.setAdditionalItems(items)
+                    self.combo.setCurrentIndex(self.combo.findText(value))
         elif self.dialogType == DIALOG_BATCH:
             self.widget.setValue(value)
         else:
             self.setComboValue(value, combobox=self.combo)
+        self.widgetValueHasChanged.emit(self)
 
     def value(self):
         if self.dialogType == DIALOG_STANDARD:
@@ -1071,10 +1187,16 @@ class FeatureSourceWidgetWrapper(WidgetWrapper):
             try:
                 layer = self.combo.currentLayer()
                 if layer is not None:
-                    return QgsProcessingFeatureSourceDefinition(layer.id(), use_selected_features)
+                    if use_selected_features:
+                        return QgsProcessingFeatureSourceDefinition(layer.id(), True)
+                    else:
+                        return layer.id()
                 else:
                     if self.combo.currentText():
-                        return QgsProcessingFeatureSourceDefinition(self.combo.currentText(), use_selected_features)
+                        if use_selected_features:
+                            return QgsProcessingFeatureSourceDefinition(self.combo.currentText(), True)
+                        else:
+                            return self.combo.currentText()
                     else:
                         return None
             except:
@@ -1110,7 +1232,7 @@ class StringWidgetWrapper(WidgetWrapper):
         else:
             # strings, numbers, files and table fields are all allowed input types
             strings = self.dialog.getAvailableValuesOfType(
-                [QgsProcessingParameterString, QgsProcessingParameterNumber, QgsProcessingParameterFile,
+                [QgsProcessingParameterString, QgsProcessingParameterNumber, QgsProcessingParameterDistance, QgsProcessingParameterFile,
                  QgsProcessingParameterField, QgsProcessingParameterExpression],
                 [QgsProcessingOutputString, QgsProcessingOutputFile])
             options = [(self.dialog.resolveValueDescription(s), s) for s in strings]
@@ -1191,6 +1313,10 @@ class StringWidgetWrapper(WidgetWrapper):
 
 class ExpressionWidgetWrapper(WidgetWrapper):
 
+    def __init__(self, param, dialog, row=0, col=0, **kwargs):
+        super().__init__(param, dialog, row, col, **kwargs)
+        self.context = dataobjects.createContext()
+
     def createWidget(self):
         if self.dialogType in (DIALOG_STANDARD, DIALOG_BATCH):
             if self.param.parentLayerParameterName():
@@ -1201,7 +1327,7 @@ class ExpressionWidgetWrapper(WidgetWrapper):
                 widget.setExpression(self.param.defaultValue())
         else:
             strings = self.dialog.getAvailableValuesOfType(
-                [QgsProcessingParameterExpression, QgsProcessingParameterString, QgsProcessingParameterNumber],
+                [QgsProcessingParameterExpression, QgsProcessingParameterString, QgsProcessingParameterNumber, QgsProcessingParameterDistance],
                 (QgsProcessingOutputString, QgsProcessingOutputNumber))
             options = [(self.dialog.resolveValueDescription(s), s) for s in strings]
             widget = QComboBox()
@@ -1223,11 +1349,10 @@ class ExpressionWidgetWrapper(WidgetWrapper):
         self.setLayer(wrapper.value())
 
     def setLayer(self, layer):
-        context = dataobjects.createContext()
         if isinstance(layer, QgsProcessingFeatureSourceDefinition):
-            layer, ok = layer.source.valueAsString(context.expressionContext())
+            layer, ok = layer.source.valueAsString(self.context.expressionContext())
         if isinstance(layer, str):
-            layer = QgsProcessingUtils.mapLayerFromString(layer, context)
+            layer = QgsProcessingUtils.mapLayerFromString(layer, self.context)
         self.widget.setLayer(layer)
 
     def setValue(self, value):
@@ -1344,23 +1469,36 @@ class VectorLayerWidgetWrapper(WidgetWrapper):
                 self.combo.setCurrentIndex(self.combo.findText(filename))
             else:
                 self.combo.setEditText(filename)
+            self.widgetValueHasChanged.emit(self)
 
     def setValue(self, value):
         if value is None or value == NULL:
             return
 
         if self.dialogType == DIALOG_STANDARD:
-            if self.combo.findText(value) >= 0:
-                self.combo.setCurrentIndex(self.combo.findText(value))
-            else:
-                items = self.combo.additionalItems()
-                items.append(value)
-                self.combo.setAdditionalItems(items)
-                self.combo.setCurrentIndex(self.combo.findText(value))
+            if isinstance(value, str):
+                layer = QgsProject.instance().mapLayer(value)
+                if layer is not None:
+                    value = layer
+
+            found = False
+            if isinstance(value, QgsMapLayer):
+                self.combo.setLayer(value)
+                found = self.combo.currentIndex() != -1
+
+            if not found:
+                if self.combo.findText(value) >= 0:
+                    self.combo.setCurrentIndex(self.combo.findText(value))
+                else:
+                    items = self.combo.additionalItems()
+                    items.append(value)
+                    self.combo.setAdditionalItems(items)
+                    self.combo.setCurrentIndex(self.combo.findText(value))
         elif self.dialogType == DIALOG_BATCH:
-            return self.widget.setText(value)
+            return self.widget.setValue(value)
         else:
             self.setComboValue(value, combobox=self.combo)
+        self.widgetValueHasChanged.emit(self)
 
     def value(self):
         if self.dialogType == DIALOG_STANDARD:
@@ -1384,6 +1522,10 @@ class VectorLayerWidgetWrapper(WidgetWrapper):
 class TableFieldWidgetWrapper(WidgetWrapper):
     NOT_SET = '[Not set]'
 
+    def __init__(self, param, dialog, row=0, col=0, **kwargs):
+        super().__init__(param, dialog, row, col, **kwargs)
+        self.context = dataobjects.createContext()
+
     def createWidget(self):
         self._layer = None
 
@@ -1427,11 +1569,10 @@ class TableFieldWidgetWrapper(WidgetWrapper):
         self.setLayer(wrapper.value())
 
     def setLayer(self, layer):
-        context = dataobjects.createContext()
         if isinstance(layer, QgsProcessingFeatureSourceDefinition):
-            layer, ok = layer.source.valueAsString(context.expressionContext())
+            layer, ok = layer.source.valueAsString(self.context.expressionContext())
         if isinstance(layer, str):
-            layer = QgsProcessingUtils.mapLayerFromString(layer, context)
+            layer = QgsProcessingUtils.mapLayerFromString(layer, self.context)
         self._layer = layer
         self.refreshItems()
 
@@ -1506,6 +1647,10 @@ class TableFieldWidgetWrapper(WidgetWrapper):
 class BandWidgetWrapper(WidgetWrapper):
     NOT_SET = '[Not set]'
 
+    def __init__(self, param, dialog, row=0, col=0, **kwargs):
+        super().__init__(param, dialog, row, col, **kwargs)
+        self.context = dataobjects.createContext()
+
     def createWidget(self):
         self._layer = None
 
@@ -1517,7 +1662,7 @@ class BandWidgetWrapper(WidgetWrapper):
         else:
             widget = QComboBox()
             widget.setEditable(True)
-            fields = self.dialog.getAvailableValuesOfType([QgsProcessingParameterBand, QgsProcessingParameterNumber],
+            fields = self.dialog.getAvailableValuesOfType([QgsProcessingParameterBand, QgsProcessingParameterDistance, QgsProcessingParameterNumber],
                                                           [QgsProcessingOutputNumber])
             if self.param.flags() & QgsProcessingParameterDefinition.FlagOptional:
                 widget.addItem(self.NOT_SET, self.NOT_SET_OPTION)
@@ -1537,11 +1682,10 @@ class BandWidgetWrapper(WidgetWrapper):
         self.setLayer(wrapper.value())
 
     def setLayer(self, layer):
-        context = dataobjects.createContext()
         if isinstance(layer, QgsProcessingParameterRasterLayer):
-            layer, ok = layer.source.valueAsString(context.expressionContext())
+            layer, ok = layer.source.valueAsString(self.context.expressionContext())
         if isinstance(layer, str):
-            layer = QgsProcessingUtils.mapLayerFromString(layer, context)
+            layer = QgsProcessingUtils.mapLayerFromString(layer, self.context)
         self._layer = layer
         self.refreshItems()
 
@@ -1621,6 +1765,8 @@ class WidgetWrapperFactory:
             wrapper = MultipleLayerWidgetWrapper
         elif param.type() == 'number':
             wrapper = NumberWidgetWrapper
+        elif param.type() == 'distance':
+            wrapper = DistanceWidgetWrapper
         elif param.type() == 'raster':
             wrapper = RasterWidgetWrapper
         elif param.type() == 'enum':
